;; Generated package description from electric-spacing.el  -*- no-byte-compile: t -*-
(define-package "electric-spacing" "5.0.0.20201201.204407" "Insert operators with surrounding spaces smartly" 'nil :url "https://elpa.gnu.org/packages/electric-spacing.html" :authors '(("William Xu" . "william.xwl@gmail.com")) :maintainer '("William Xu" . "william.xwl@gmail.com"))
